from macrolib.typemacros import tupler, dict_union, replace_value_nested



#----------------------------------------------------------------------------------------------------------------------
#Result Class

class Result():
    def __init__(self, n: int):
        self.n = n
    def __getitem__(self, n: int):
        return Result(n)
    def __repr__(self):
        return f"<class Result[{self.n}]>"
    def __eq__(self, other):
        return isinstance(other, Result) and self.n == other.n


#----------------------------------------------------------------------------------------------------------------------
#Menu Class

class Menu():
    #Keyword Objects
    result = Result(-1)
    escape = object()
    __END__ = type("Menu.__END__", (object,), {})  #Terminal Object

    #Matching Keywords
    exit_to = object()  #Use as keyword for end_to to match exit_to:  Menu(end_to = Menu.exit_to)
    end_to = object()

    #Kwargs Wrapper
    class kwargs(dict):
        pass


    def __init__(self,
                 name = "Choose Action",
                 exit_to = lambda: Menu.__END__,
                 end_to = None,
                 escape_to = None,
                 exit_key = "e",
                 exit_message = "exit",
                 empty_message = "--*No Entries*--"
                 ):
        #Pass parameters
        self.name = name
        self.exit_to = exit_to
        self.empty_message = empty_message

        #Break point matching heirarchy
        self.end_to = self if end_to is None else \
            (self.exit_to if end_to is Menu.exit_to else end_to)
        self.escape_to = self if escape_to is None else \
            (self.exit_to if escape_to is Menu.exit_to else
             (self.end_to if escape_to is Menu.end_to else escape_to))

        #Init menu data
        self.menu_display_list = []      #["[key]- message"]
        self.menu = {}                   #{"key":(function-arg chain)}

        #Set exit option
        self.exit = (exit_key, exit_message, (exit_to, ()))
        self.update_menu(self.exit)


    def __call__(self):
        """Runs menu and entry function chain"""
        #Exit On Empty Menu
        if not self.menu_display_list[:-1]:
            print(self.empty_message)
            return self.exit_to()

        #Print Menu And Get Input
        show = f"\n{self.name}\n" + "\n".join(self.menu_display_list) + "\n"
        selection = input(show)

        #Refresh Menu On Invalid Input
        if selection not in self.menu.keys():
            print("--*Invalid Selection*--")
            return self()

        #Select Item
        switch = self.menu[selection]

        #Evaluate Function Chain
        result = None
        results = []
        while len(switch) >= 2:
            #Get func/args pair
            func = switch[0]
            args = tupler(switch[1])

            #Replace Result Keywords
            N = len(results)
            for n in range(-N, N):
                func = replace_value_nested(tupler(func), Result(n), results[n])[0]
                args = replace_value_nested(tupler(args), Result(n), results[n])

            #Separate args/kwargs
            kwargs = dict_union(tupler(x for x in args if isinstance(x, Menu.kwargs)))
            args = tupler(x for x in args if not isinstance(x, Menu.kwargs))

            #Evaluate Function
            result = Bind.lazy_eval(func, args, kwargs)

            #Manual escape
            if result is Menu.escape:
                return self.escape_to()

            #End Loop
            results.append(result)
            switch = switch[2:]

        #Go to end_to if last result is None
        return result if result is not None else self.end_to()


    def append(self,*data):
        """
        Append data to menu in the form:
        ('key', 'message', (func1, (*args1), func2, (*args2),...))
        --*{forces exit key to the end of the list}*--
        """
        self.menu_display_list = self.menu_display_list[:-1]
        for n in data + (self.exit,):
            self.update_menu(n)


    def update_menu(self, data):
        """Updates menu lists"""
        self.menu_display_list.append(f"[{data[0]}]- {data[1]}")
        self.menu[data[0]] = data[2]



#----------------------------------------------------------------------------------------------------------------------
#Lazy Evaluation

class Bind():
    """
    Used for delayed evaluation.
    Creates a callable list of function / argument pairs in the form:
    [<function>, (args), {kwargs}]  ->  function(*args, **kwargs)
    """
    class Wrapper(list):
        def __call__(self):
            return Bind.lazy_eval(self[0], self[1], self[2])


    def __new__(cls, func, *args, **kwargs) -> Wrapper:
        return cls.Wrapper([func, args, kwargs])


    @staticmethod
    def lazy_eval(func, args, kwargs):
        """Depth-first evaluation of nested function/argument bindings."""
        func = func() if isinstance(func, Bind.Wrapper) else func
        args = tupler(x() if isinstance(x, Bind.Wrapper) else x for x in args)
        kwargs = {k: v() if isinstance(v, Bind.Wrapper) else v for k, v in kwargs.items()}

        return func(*args, **kwargs)



#----------------------------------------------------------------------------------------------------------------------
#In-Line Functions

def escape_on(x, value):
    """Returns an escape if the two arguments are equal, or both Truthy or both Falsy.
    Otherwise, returns value."""
    return Menu.escape if x == value or bool(x) == bool(value) else value


def f_escape(*args, **kwargs) -> Menu.escape:
    """Polymorphic in-line escape function."""   #terminal morphism in Hom(*,escape)
    return Menu.escape


def f_switch(n: int, funcs: tuple) -> Bind.Wrapper:
    """Returns a lazy function of type (int -> function)"""
    return Bind(lambda b: funcs[b], n)



#----------------------------------------------------------------------------------------------------------------------
#Builtin Menus

def yesno_ver(**kwargs) -> bool:
    """Simple yes/no verification returning bool"""
    kwargs_ = {"name":"Are you sure?", "exit_message":"cancel"} | kwargs | {"exit_to":lambda: False}
    menu = Menu(**kwargs_)
    menu.append(("x", "yes", (lambda: True, ())))

    return menu()


def edit_list(entries: list | tuple, **kwargs) -> list | tuple:
    """Delete items in a list/tuple; returns updated list/tuple"""
    kwargs_ = {"name":"Edit List"} | kwargs | {"exit_to":lambda: entries}
    menu = Menu(**kwargs_)
    for n, entry in enumerate(entries):
        menu.append((str(n), str(entry), (edit_list, (entries[:n] + entries[n+1:], Menu.kwargs(kwargs)))))

    return menu()



from lib.menucmd import Menu, edit_list, yesno_ver, f_escape, f_switch
from lib.menucmd import Bind as B
import numpy as np


def main():
    result = Menu.result
    kwargs = Menu.kwargs

    main_menu = Menu(name = "Hub", exit_key = "x")
    lazy_menu = Menu(name = "Lazy Eval", exit_to = main_menu)
    builtin_menu = Menu(name = "Builtins", exit_to = main_menu)


    main_menu.append(
        ("e", "Test Lazy Eval", (lazy_menu, (),)),
        ("b", "Test Builtins", (builtin_menu, (),)),
    )
#
    lazy_menu.append(
        ("s", "Square Number", (
            input, ("Number: "),
            lambda x: x**2, B(float, result),
            print, (result),
        )),

        ("c", "Cards", (
            shuffle_cards, (),
            pick_card, (B(int,B(input, "Choose Card (0-51): ")), result),
            print, (result),
        )),

        ("r", "Result Memory", (
            input, "++n: ", int, result, print, "Partial Sums:"
            ) + sum([(sum, [result[-2*j-2] for j in range(u+1)], print, result) for u in range(10)], start = ())
        )

    )
#
    builtin_menu.append(
        ("l", "Edit List", (
            input, ("List items (separated by spaces): "),
            edit_list, (B(lambda s: s.split(), result), kwargs(name = "Delete Items")),
            print, ("New string:\n", B(" ".join, result)),
        )),

        ("v", "Confirmation", (
            yesno_ver, kwargs(name = "Delete system32?", exit_message = "no"),
            f_switch(result, (f_escape, yesno_ver)), kwargs(name ="Are you really sure?", exit_message = "no"),
            f_switch(result, (f_escape, yesno_ver)), kwargs(name = "Are you not sure?", exit_message = "no"),
            f_switch(result, (yesno_ver, f_escape)), kwargs(name = "Are you sure you are not sure?", exit_message = "wtf?"),
            print, B(lambda b: "crisis averted!" if b else "deleting...\nsystem32 cannot be deleted!", result)
        )),
    )

    main_menu()


#---------------------------------------------------------------------------------------------------------------

def shuffle_cards() -> np.array:
    suites = np.array([x for x in range(4)])
    numbers = np.array([x for x in range(13)])
    deck = np.array(np.meshgrid(suites, numbers)).T.reshape(-1, 2)
    np.random.shuffle(deck)

    return deck


def pick_card(n: int, deck: np.array) -> str:
    display_suite = {0:"Clubs", 1:"Spades", 2:"Hearts", 3:"Diamonds"}
    display_number = {0:"Ace"} | {v:str(v+1) for v in range(1,9)} | {10:"Jack", 11:"Queen", 12:"King"}

    return display_number[deck[n][1]] + " of " + display_suite[deck[n][0]]


#------------------------------------------------------------------------------------------------------------

def get_code() -> list[str]:
    with open("menutesting.py", "r") as code:
        lines = code.readlines()
    code.close()

    return lines



#################################################################################################################

if __name__ == "__main__":
    main()


from types import GeneratorType
from typing import Any


#Unions
#----------------------------------------------------------------------------------------------------------------------

def list_union(a: list | tuple | list[list] | tuple[list], b: list | tuple | None = None) -> list:
    """
    If 'b' is None, the function will perform an iterative union of all lists and tuples in 'a'.
    In this case, 'a' must be a list or tuple of exclusively lists or  tuples.
    Otherwise, perform the union of 'a' and 'b'
    :param a: list or tuple (of lists or tuples)
    :param b: list or tuple
    :return: list
    """
    if b is None:
        l_union = list()
        for n in a:
            if isinstance(n, list | tuple):
                l_union = list_union(n, l_union)
            else:
                print("'a' must be a list or tuple of lists or tuples if 'b' is empty")
                raise TypeError
        return l_union
    else:
        return list(set(a) | set(b))


def tuple_union(a: list | tuple | list[tuple] | tuple[tuple], b: list | tuple | None = None) -> tuple:
    """
    If 'b' is None, the function will perform an iterative union of all lists and tuples in 'a'.
    In this case, 'a' must be a list or tuple of exclusively lists or  tuples.
    Otherwise, perform the union of 'a' and 'b'
    :param a: list or tuple (of lists or tuples)
    :param b: list or tuple
    :return: tuple
    """
    if b is None:
        t_union = tuple()
        for n in a:
            if isinstance(n, list | tuple):
                t_union = tuple_union(n, t_union)
            else:
                print("'a' must be a list or tuple of lists or tuples if 'b' is empty")
                raise TypeError
        return t_union
    else:
        return tuple(set(a) | set(b))


def dict_union(a: dict | list[dict] | tuple[dict], b: dict | None = None) -> dict:
    """
    If 'b' is None, the function will perform an iterative union of all dicts in 'a'.
    In this case, 'a' must be a list or tuple of exclusively dicts.
    Otherwise, perform the union of 'a' and 'b'
    (Leftmost keys have precedence)
    :param a: dict, list, or tuple (of dicts)
    :param b: set
    :return: dict
    """
    if b is None:
        d_union = dict()
        for n in a:
            if isinstance(n, dict):
                d_union = dict_union(n, d_union) #test switching these around
            else:
                print("'a' must be a list or tuple of dicts if 'b' is empty")
                raise TypeError
        return d_union
    else:
        return dict(a | b)


def set_union(a: set | list[set] | tuple[set], b: set | None = None) -> set:
    """
    If 'b' is None, the function will perform an iterative union of all sets in 'a'.
    In this case, 'a' must be a list or tuple of exclusively sets.
    Otherwise, perform the union of 'a' and 'b'
    :param a: set, list, or tuple (of dicts)
    :param b: set
    :return: set
    """
    if b is None:
        s_union = set()
        for n in a:
            if isinstance(n, set):
                s_union = set_union(n, s_union)
            else:
                print("'a' must be a list or tuple of sets if 'b' is empty")
                raise TypeError
        return s_union
    else:
        return set(a | b)


#Syntax Converters
#----------------------------------------------------------------------------------------------------------------------

def tupler(a) -> tuple:
    """
    Passes tuples and generator objects to tuples, and parenthesizes anything else.
    This effectively bypasses the need for a comma in defining a singleton tuple;  (a) -> (a,)
    """
    if isinstance(a, tuple | GeneratorType):
        return tuple(a)
    else:
        return (a,)


#Expansions
#----------------------------------------------------------------------------------------------------------------------

def replace_value_nested(a: list | tuple | dict | set, old_vals: tuple | Any , new_val) -> list | tuple | dict | set:
    """
    Replaces a value recursively in a data structure.  The value(s) to replace can be of any type,
    including any type of the data structures being searched through.
    :param a: list, tuple, dict, or set  (all nestings allowed)
    :param old_vals: value(s) to replace
    :param new_val: value to inject
    :return: list, tuple, dict, or set
    """
    old_vals = tupler(old_vals)

    if a in old_vals:
        return new_val
    elif isinstance(a, list):
        return type(a)(replace_value_nested(x, old_vals, new_val) for x in a)
    elif isinstance(a, tuple):
        return type(a)(replace_value_nested(x, old_vals, new_val) for x in a)
    elif isinstance(a, dict):
        return type(a)({k:replace_value_nested(v, old_vals, new_val) for k,v in a.items()})
    elif isinstance(a, set):
        return type(a)(replace_value_nested(x, old_vals, new_val) for x in a)
    else:
        return new_val if a in old_vals else a




from setuptools import setup, find_packages

setup(
    name="menucmd",
    version="1.00.1",
    packages=find_packages(),
    install_requires=[
        # List your dependencies here
    ],
    author="Casey Litmer",
    author_email="litmerc@msn.com",
    description="Command line menu interface",
    long_description=open('README.md').read(),
    long_description_content_type='text/markdown',
    url="https://github.com/Casey-Litmer/menucmd",
    classifiers=[
        "Programming Language :: Python :: 3",
        "Operating System :: OS Independent",
    ],
    python_requires='>=3.6',
)
